package com.artificialsolutions.asolfbconnector.bridge;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.artificialsolutions.asolfbconnector.common.Config;
import com.artificialsolutions.asolfbconnector.facebook.FacebookMessagingClient;
import com.artificialsolutions.asolfbconnector.teneoengine.TeneoEngineClient;
import com.artificialsolutions.common.OneTimeTask;

import jakarta.json.JsonNumber;
import jakarta.json.JsonObject;
import jakarta.json.JsonString;
import jakarta.json.JsonValue;

/**
 * A bridge communicator sending user inputs to a Teneo engine and submitting the received Teneo
 * engine responses to Facebook Messenger. It manages the sessions both for the communication with
 * Facebook Messenger and for Teneo engine. The application session is based on the user's Facebook
 * page-scoped ID, while the Teneo engine session is handled via cookie in the standard Java web
 * application way. The instances of this class are thread-safe and Serializable.
 */
public class FacebookTeneoBridge implements Serializable {

    /**
     * Serial version ID for object (de)serialization.
     */
    private static final long serialVersionUID = 1944356823720366109L;

    /**
     * An application session implementation. Each session contains a Teneo engine client, a Facebook
     * messaging client and a session timeout task to terminate and eliminate the session after a period
     * of inactivity.
     */
    final class Session implements Serializable {

        /**
         * Serial version ID for object (de)serialization.
         */
        private static final long serialVersionUID = 4718586634808533996L;

        /**
         * The user's Facebook page-scoped ID.
         */
        final String userPsid;

        /**
         * Teneo engine client.
         */
        final TeneoEngineClient teneoEngineClient;

        /**
         * Facebook messaging client.
         */
        final FacebookMessagingClient facebookMessagingClient;

        /**
         * Flag indicates if the session is expired ({@code true}) or still alive ({@code false}).
         */
        transient volatile boolean expired;

        /**
         * The time point in milliseconds as per {@link java.lang.System#currentTimeMillis()
         * System.currentTimeMillis()} when the session timeout task was started (or restarted).
         */
        transient volatile long timeoutTaskStart = -1;

        /**
         * The current session timeout task. Its goal is to terminate the session, which includes closing
         * all the communication clients the session object contains and deleting the session object itself
         * from the session repository.
         */
        transient volatile OneTimeTask timeoutTask;

        /**
         * Constructs a session object.
         * 
         * @param userPsid the user's Facebook page-scoped ID the session is created for. It is used as a
         * session ID of this communicator.
         * @param pageAccessToken the Facebook page access token of the business page whose messenger serves
         * as a frontend to Teneo engine.
         */
        Session(final String userPsid, final String pageAccessToken) {
            this.userPsid = userPsid;
            this.teneoEngineClient = new TeneoEngineClient(config);
            this.facebookMessagingClient = new FacebookMessagingClient(config, userPsid, pageAccessToken);
        }

        /**
         * Returns the user's Facebook page-scoped ID this session is associated with.
         * 
         * @return the user's Facebook page-scoped ID.
         */
        String getUserPsid() {
            return userPsid;
        }

        /**
         * Returns the Teneo engine client for this session.
         * 
         * @return this session's Teneo engine client.
         */
        TeneoEngineClient getTeneoEngineClient() {
            return teneoEngineClient;
        }

        /**
         * Returns the Facebook messaging client for this session.
         * 
         * @return this session's Facebook messaging client.
         */
        FacebookMessagingClient getFacebookMessagingClient() {
            return facebookMessagingClient;
        }

        /**
         * Indicates if the session has expired.
         * 
         * @return the state of the session.
         */
        synchronized boolean isExpired() {
            return expired;
        }

        /**
         * Returns the time point in milliseconds (as per {@link java.lang.System#currentTimeMillis()
         * System.currentTimeMillis()}) when the session timeout task was started (or restarted). This value
         * will be {@code -1} if either the timer has not been started yet or the session expired.
         * 
         * @return the time point in milliseconds.
         */
        synchronized long getTimeoutTaskStart() {
            return timeoutTaskStart;
        }

        /**
         * Cancels the scheduled timeout task.
         */
        synchronized void clearTimeoutTask() {
            if (timeoutTask == null) return;
            timeoutTask.cancel();
            timeoutTask = null;
        }

        /**
         * "Expires" the session: the eventual pending timeout task is cancelled, the timing start value is
         * reset and the session object is removed from the session repository. If this method is called for
         * an already expired session, it does nothing.
         */
        void expire() {
            final int sessionCount;
            synchronized (userPsidToSession) {
                synchronized (this) {
                    clearTimeoutTask();
                    expired = true;
                    timeoutTaskStart = -1;
                    userPsidToSession.remove(userPsid);
                    // This here guarantees there are no STATE_EXPIRED sessions in the repo
                    // accessed as synchronized
                }
                sessionCount = userPsidToSession.size();
            }
            if (logger.isDebugEnabled()) {
                if (logSensitive) logger.debug("Expiring and removing session with id/userPsid [{}]; simultaneous sessions: {}", userPsid, sessionCount);
                else logger.debug("Expiring and removing session; simultaneous sessions: {}", sessionCount);
            }
        }

        /**
         * (Re)starts the session timeout count-down by canceling the current scheduled timeout task (if
         * there is one) and scheduling a new one. If this method is called for an already expired session,
         * it does nothing.
         * 
         * @param timeoutMillis the number of milliseconds after which the session expires.
         * @param resumeAfterPause indicates if it is a count-down resumption ({@code true}) after the
         * timeout task was temporarily paused or a normal timer (re)start ({@code false}).
         */
        void handleTiming(final long timeoutMillis, final boolean resumeAfterPause) {
            boolean bExpired = false;
            synchronized (this) {
                clearTimeoutTask();
                if (expired) bExpired = true;
                else {
                    // The session has not expired yet. Schedule a (new) session timeout task:
                    timeoutTask = new OneTimeTask(timeoutMillis) {
                        @Override
                        public void run() {
                            expire();
                        }
                    };
                    if (resumeAfterPause) timeoutTask.start();
                    else timeoutTaskStart = timeoutTask.start();
                }
            }
            if (bExpired) {
                if (logSensitive) logger.info("Hitting upon an expired session with id/userPsid [{}]", userPsid);
                else logger.info("Hitting upon an expired session");
            } else {
                if (logSensitive) logger.debug("Scheduling a new timeout task for session with id/userPsid [{}]", userPsid);
                else logger.debug("Scheduling a new timeout task for session");
            }
        }

        /**
         * Saves this object to a stream (i.e., serializes it).
         *
         * @serialData the timeout flag {@link #expired} and the time point of the session timeout task
         * start {@link #timeoutTaskStart} are serialized in a controlled way to guarantee synchronization.
         *
         * @param oos the stream.
         * 
         * @throws IOException if an I/O error occurs.
         */
        private void writeObject(final ObjectOutputStream oos) throws IOException {
            oos.defaultWriteObject();
            synchronized (this) {
                oos.writeBoolean(expired);
                oos.writeLong(timeoutTaskStart);
            }
        }

        /**
         * Reconstitutes this object from a stream (i.e., deserializes it).
         * 
         * @param ois the stream.
         * 
         * @throws IOException if the class of a serialized object could not be found.
         * @throws ClassNotFoundException if an I/O error occurs.
         */
        private void readObject(final ObjectInputStream ois) throws IOException, ClassNotFoundException {
            ois.defaultReadObject();
            expired = ois.readBoolean();
            timeoutTaskStart = ois.readLong();
        }
    }

    /**
     * The configuration object.
     */
    private final Config config;

    /**
     * Session repository mapping a session ID to the corresponding session object.
     */
    private transient volatile Map<String, Session> userPsidToSession;

    /**
     * The logger.
     */
    private transient Logger logger;

    /**
     * Indicates if potentially GDPR-sensitive data may ({@code true}) or may not ({@code false}) be
     * logged. The value of this property should be {@code false} in production.
     */
    private transient boolean logSensitive;

    /**
     * Constructs an instance of this class.
     * 
     * @param config the configuration object.
     */
    public FacebookTeneoBridge(final Config config) {
        this.config = config;
        userPsidToSession = new HashMap<>();
        initLogging();
        logger.trace("An instance of {} is constructed", FacebookTeneoBridge.class);
    }

    /**
     * Initializes the logging.
     */
    private void initLogging() {
        logger = LoggerFactory.getLogger(FacebookTeneoBridge.class);
        logSensitive = logger.isDebugEnabled() || logger.isTraceEnabled();
    }

    /**
     * Provides a {@link Session} object for the given ID. If there already exists a {@link Session}
     * object for this ID, its timeout task is restarted and it is returned. If not, a new one is
     * created and associated with this ID, its timeout task is started and it is returned.
     * 
     * @param userPsid the ID (the Facebook Messenger user's page-scoped ID) of the requested session.
     * @param pageAccessToken the Facebook page access token of the business page whose messenger serves
     * as a frontend to Teneo engine.
     * 
     * @return the {@link Session} object or {@code null} if <i>too many</i> parallel sessions exist
     * (see {@link Config#PROP_BRIDGE_MAX_PARALLEL_SESSIONS} end
     * {@link Config#getMaxParallelSessions()}).
     */
    private Session provideSession(final String userPsid, final String pageAccessToken) {
        Session session;
        final int sessionCount;
        boolean newSession = false;
        synchronized (userPsidToSession) {
            session = userPsidToSession.get(userPsid);
            if (session == null && userPsidToSession.size() < config.getMaxParallelSessions()) {
                newSession = true;
                userPsidToSession.put(userPsid, session = new Session(userPsid, pageAccessToken));
            }
            sessionCount = userPsidToSession.size();
            if (session != null) session.handleTiming(config.getBridgeSessionTimeoutMillis(), false);
        }
        if (session == null) logger.warn("Failing to add new session because the session limit {} is exceeded", config.getMaxParallelSessions());
        else if (logger.isDebugEnabled()) {
            if (newSession) {
                if (logSensitive) {
                    logger.debug("Adding new session with id/userPsid [{}]; simultaneous sessions: {}", userPsid, sessionCount);
                } else {
                    logger.debug("Adding new session; simultaneous sessions: {}", sessionCount);
                }
            } else {
                if (logSensitive) {
                    logger.debug("Getting existing session with id/userPsid [{}]; simultaneous sessions: {}", userPsid, sessionCount);
                } else {
                    logger.debug("Getting existing session; simultaneous sessions: {}", sessionCount);
                }
            }
        }
        return session;
    }

    /**
     * Asynchronously submits a user input to Teneo engine and then sends the Teneo engine response to
     * Facebook Messenger. In this way this response is displayed to the user currently talking on the
     * messenger to the actual business page. The user is identified via his/her page-scoped ID (PSID)
     * and the page is identified via its page access token. This method works in the same way as
     * {@link #submitTeneoFacebook(String, String, Map)} with the only difference that the underlying
     * Teneo request here is asynchronous and it does not wait for Teneo to respond.
     *
     * @param userPsid the user's page-scoped ID (PSID).
     * @param pageAccessToken the page access token.
     * @param paramNameToValue data submitted to Teneo engine - a mapping of request parameter names to
     * their values.
     */
    public void submitTeneoFacebookAsync(final String userPsid, final String pageAccessToken, final Map<String, ?> paramNameToValue) {
        final Session session = provideSession(userPsid, pageAccessToken);
        if (session != null) {
            session.getTeneoEngineClient().sendAsync(paramNameToValue).whenComplete((final JsonObject json, final Throwable thr) -> {
                doAfterTeneoEngineQuery(session, json, thr);
            });
        } else {
            // Concurrent session limit has been surpassed. This souldn't happen in normal cases.
            new FacebookMessagingClient(config, userPsid, pageAccessToken).sendTextAsync("Session limit for Teneo Facebook bridge has been reached");
        }
    }

    /**
     * Submits a user input to Teneo engine and then sends the Teneo engine response to Facebook
     * Messenger. In this way this response is displayed to the user currently talking on the Facebook
     * messenger to the actual business page. The user is identified via his/her page-scoped ID (PSID)
     * and the page is identified via its page access token. This method returns after the Teneo engine
     * response is received, but does not wait for the response of the Facebook messenger.
     * 
     * @see #submitTeneoFacebookAsync(String, String, Map)
     * 
     * @param userPsid the user's page-scoped ID (PSID).
     * @param pageAccessToken the page access token.
     * @param paramNameToValue data submitted to Teneo engine - a mapping of request parameter names to
     * their values.
     * 
     * @return {@code true} if the there have been no error and the session is still alive;
     * {@code false} otherwise.
     */
    public boolean submitTeneoFacebook(final String userPsid, final String pageAccessToken, final Map<String, ?> paramNameToValue) {
        final Session session = provideSession(userPsid, pageAccessToken);
        if (session == null) {
            new FacebookMessagingClient(config, userPsid, pageAccessToken).sendTextAsync("Session limit for Teneo Facebook bridge has been reached");
            return false;
        }
        JsonObject json = null;
        Exception thr = null;
        try {
            json = session.getTeneoEngineClient().send(paramNameToValue);
        } catch (final Exception ex) {
            thr = ex;
        }
        doAfterTeneoEngineQuery(session, json, thr);
        return (thr == null);
    }

    /**
     * Processes an eventual error or submits data to Teneo engine. If the session expires in the
     * meantime, it also closes the Teneo engine session.
     * 
     * @param session the session.
     * @param json the object containing data to submit to Teneo engine.
     * @param thr the error.
     */
    private void doAfterTeneoEngineQuery(final Session session, final JsonObject json, final Throwable thr) {
        if (thr != null) {
            logger.error("Teneo response failure", thr);
            if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync("Teneo response failure: " + thr);
            return;
        }
        String s;

        JsonValue jv = json.get("status");
        if (jv == null) {
            s = "Teneo response has no [status] property";
            if (logSensitive) s += "; response: " + json;
            logger.error(s);
            if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
        } else if (jv.getValueType() != JsonValue.ValueType.NUMBER) {
            s = "Teneo response has [status] of type " + jv.getValueType() + ", should be " + JsonValue.ValueType.NUMBER;
            if (logSensitive) s += "; response: " + json;
            logger.error(s);
            if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
        } else {
            switch (((JsonNumber)jv).intValue()) {
                case 0:
                    // OK - Teneo engine answered
                    final JsonValue output = json.get("output");
                    if (output == null) {
                        s = "Teneo response has no [output] property";
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else if (output.getValueType() != JsonValue.ValueType.OBJECT) {
                        s = "Teneo response has [output] of type " + output.getValueType() + ", should be " + JsonValue.ValueType.OBJECT;
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else if ((jv = output.asJsonObject().get("parameters")) != null && jv.getValueType() == JsonValue.ValueType.OBJECT && (jv = jv.asJsonObject().get("facebookMessage")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                        switch (jv.getValueType()) {
                            case STRING:
                                // Parse
                                s = ((JsonString)jv).getString();
                                if (logSensitive) logger.debug("Submitting Teneo facebookMessage response to Facebook: [{}]", s);
                                else logger.debug("Submitting Teneo facebookMessage response to Facebook");
                                session.getFacebookMessagingClient().sendAsync(s);
                                break;
                            case OBJECT:
                                if (logSensitive) logger.debug("Submitting Teneo facebookMessage response to Facebook: [{}]", jv);
                                else logger.debug("Submitting Teneo facebookMessage response to Facebook");
                                session.getFacebookMessagingClient().sendAsync((JsonObject)jv);
                                break;
                            default:
                                s = "Teneo response has [output.parameters.facebookMessage] of type " + jv.getValueType() + ", should be " + JsonValue.ValueType.STRING + " or " + JsonValue.ValueType.OBJECT;
                                if (logSensitive) s += "; response: " + json;
                                logger.error(s);
                                if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                        }
                    } else if ((jv = output.asJsonObject().get("text")) == null) {
                        s = "Teneo response has no [output.text]";
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else if (jv.getValueType() != JsonValue.ValueType.STRING) {
                        s = "Teneo response has [output.text] of type " + jv.getValueType() + ", should be " + JsonValue.ValueType.STRING;
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else {
                        s = ((JsonString)jv).getString();
                        if (logSensitive) logger.debug("Submitting Teneo text response to Facebook: [{}]", s);
                        else logger.debug("Submitting Teneo rtext esponse to Facebook");
                        session.getFacebookMessagingClient().sendTextAsync(s);
                    }
                    break;

                case -1:
                    // Error - Teneo engine returned an error
                    if ((jv = json.asJsonObject().get("message")) == null) {
                        s = "Teneo error response has no [message] property";
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else if (jv.getValueType() != JsonValue.ValueType.STRING) {
                        s = "Teneo error response has [message] of type " + jv.getValueType() + ", should be " + JsonValue.ValueType.STRING;
                        if (logSensitive) s += "; response: " + json;
                        logger.error(s);
                        if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
                    } else {
                        s = ((JsonString)jv).getString();
                        if (logSensitive) logger.debug("Submitting Teneo error message to Facebook: [{}]", s);
                        else logger.debug("Submitting Teneo error message to Facebook");
                        session.getFacebookMessagingClient().sendTextAsync(s);
                    }
                    break;

                default:
                    // Error - Teneo engine returned a status code unexpected in this context
                    s = "Teneo response has an unexpected value of the [status] property for a regular request: " + ((JsonNumber)jv).intValue();
                    if (logSensitive) s += "; response: " + json;
                    logger.error(s);
                    if (config.isExplicitData()) session.getFacebookMessagingClient().sendTextAsync(s);
            }
        }
        s = null;

        if (session.isExpired()) {
            if (logSensitive) logger.debug("Session was cancelled for id/userPsid [{}], forcing Teneo engine endsession", session.getUserPsid());
            else logger.debug("Session was cancelled, forcing endsession");
            try {
                session.getTeneoEngineClient().endSessionAsync();
            } catch (final Exception ex) {
                logger.warn("Failure ending Teneo session", ex);
            }
        }
    }

    /**
     * Saves this object to a stream (i.e., serializes it).
     *
     * @serialData the session repo {@link #userPsidToSession} and the sessions contained therein are
     * serialized in a controlled way to guarantee synchronization.
     * 
     * @param oos the stream.
     * 
     * @throws IOException if an I/O error occurs.
     */
    private void writeObject(final ObjectOutputStream oos) throws IOException {
        oos.defaultWriteObject();
        synchronized (userPsidToSession) {
            // All the sessions here are guaranteed to be of STATE_RUNNING
            oos.writeObject(userPsidToSession);
        }
    }

    /**
     * Reconstitutes this object from a stream (i.e., deserializes it).
     * 
     * @serialData the session repo {@link #userPsidToSession} and the sessions contained therein are
     * deserialized in a controlled way to guarantee the correct handling of the timeout tasks and
     * timeouts.
     * 
     * @param ois the stream.
     * 
     * @throws IOException if the class of a serialized object could not be found.
     * @throws ClassNotFoundException if an I/O error occurs.
     */
    private void readObject(final ObjectInputStream ois) throws IOException, ClassNotFoundException {
        ois.defaultReadObject();
        userPsidToSession = (Map<String, Session>)ois.readObject();
        synchronized (userPsidToSession) {
            // All the sessions here are guaranteed to be STATE_RUNNING with the correct timeoutTaskStart value
            for (final Iterator<Map.Entry<String, Session>> iter = userPsidToSession.entrySet().iterator(); iter.hasNext();) {
                final Session session = iter.next().getValue();
                // Calculate the new, reduced timeout as a real timeout minus
                // the time that elapsed after the timing start:
                final long reducedTimeout = config.getBridgeSessionTimeoutMillis() - (System.currentTimeMillis() - session.getTimeoutTaskStart());
                if (reducedTimeout > 10) session.handleTiming(reducedTimeout, true);
                else iter.remove();
            }
        }
        initLogging();
        logger.trace("An instance of {} is reconstituted via deserialization", FacebookTeneoBridge.class);
    }
}
