package com.artificialsolutions.asolfbconnector.common;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.function.Function;
import java.util.function.ToIntFunction;
import java.io.File;
import java.io.Reader;
import java.io.Serializable;
import java.io.StringReader;
import java.net.URI;
import java.net.URL;
import java.nio.charset.Charset;
import java.io.FileReader;
import java.io.IOException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import jakarta.json.Json;
import jakarta.json.JsonReader;
import jakarta.json.JsonString;
import jakarta.json.JsonValue;

/**
 * Global configuration object.
 */
public class Config implements Serializable {

    /**
     * Serial version ID for object (de)serialization.
     */
    private static final long serialVersionUID = 4902449067499652244L;

    /**
     * The charset used for encoding.
     */
    private static final Charset charset = Charset.forName("UTF-8");

    /**
     * <code>{@value}</code> - the property name for the URL of the Teneo engine.
     */
    public static final String PROP_TENEO_ENGINE_URL = "teneo.engine.endpointUrl";

    /**
     * <code>{@value}</code> - the property name for the connection timeout of the requests sent to the
     * Teneo engine.
     */
    public static final String PROP_TENEO_ENGINE_CONNECT_TIMEOUT = "teneo.engine.connectTimeoutMillis";

    /**
     * <code>{@value}</code> - the property name for the response timeout of the requests sent to the
     * Teneo engine.
     */
    public static final String PROP_TENEO_ENGINE_RESPONSE_TIMEOUT = "teneo.engine.responseTimeoutMillis";

    /**
     * <code>{@value}</code> - the property name for the connection timeout of the requests sent to
     * Facebook Messenger.
     */
    public static final String PROP_FACEBOOK_MESSAGING_CONNECT_TIMEOUT = "facebook.messaging.connectTimeoutMillis";

    /**
     * <code>{@value}</code> - the property name for the response timeout of the requests sent to
     * Facebook Messenger.
     */
    public static final String PROP_FACEBOOK_MESSAGING_RESPONSE_TIMEOUT = "facebook.messaging.responseTimeoutMillis";

    /**
     * <code>{@value}</code> - the property name for the session timeout in milliseconds of the
     * Facebook-Teneo bridge. For the best efficiency, this value should be some 10-20 seconds longer
     * than the timeout of the Teneo engine. Setting it longer will result in more parallel session
     * coexisting in the bridge. Setting it shorted than the Teneo engine timeout will have the same
     * effect for the user as shortening the Teneo engine session timeout itself without reducing the
     * number of parallel session in the Teneo engine though, so it should be avoided.
     */
    public static final String PROP_BRIDGE_SESSION_TIMEOUT = "bridge.sessionTimeoutMillis";

    /**
     * <code>{@value}</code> - the property name for the maximal number of parallel session in the the
     * Facebook-Teneo bridge.
     */
    public static final String PROP_BRIDGE_MAX_PARALLEL_SESSIONS = "bridge.maxParallelSessions";

    /**
     * <code>{@value}</code> - the property name for the Facebook Messenger webhook verification token
     * as per the <a href=
     * "https://developers.facebook.com/docs/workplace/integrations/custom-integrations/webhooks/">webhooks
     * documentation</a>.
     */
    public static final String PROP_FACEBOOK_WEBHOOK_VERIFICATION_TOKERN = "facebook.webhook.verificationToken";

    /**
     * <code>{@value}</code> - the property name for the Facebook Messenger application app secret to be
     * used for
     * <a href= "https://developers.facebook.com/docs/messenger-platform/webhook#security">Validating
     * Webhook Events</a>.
     */
    public static final String PROP_FACEBOOK_APPLICATION_SECRET = "facebook.application.appSecret";

    /**
     * <code>{@value}</code> - the property name for the JSON-formatted mapping from Facebook page IDs
     * to the access tokens of those pages. For example:
     * 
     * <pre>
     * facebook.pageIdToAccessToken={\
     *   "myTestPage1": "Page1Token1234567890-test",\
     *   "myTestPage2": "Page2Token0987654321-test"\
     * }
     * </pre>
     */
    public static final String PROP_FACEBOOK_PAGE_ID_TO_ACCESS_TOKEN = "facebook.pageIdToAccessToken";

    /**
     * <code>{@value}</code> - the property name for the Boolean value indicating if some error and
     * debug information should be added to requests sent both to Teneo engine and to Facebook
     * Messenger. This property is not obligatory and defaults to {@code false}. It should only be set
     * to {@code true} for testing and troubleshooting.
     */
    public static final String PROP_APPLICATION_EXPLICIT_DATA = "application.explicitData";

    /**
     * The URL (URI) of the Teneo engine fordialog requests.
     */
    private final URI teneoEngineEndpointUri;

    /**
     * The connection timeout of the requests sent to the Teneo engine.
     */
    private final int teneoEngineConnectTimeoutMillis;

    /**
     * The response timeout of the requests sent to the Teneo engine.
     */
    private final int teneoEngineResponseTimeoutMillis;

    /**
     * The connection timeout of the requests sent to Facebook Messenger.
     */
    private final int facebookMessagingConnectTimeoutMillis;

    /**
     * The response timeout of the requests sent to Facebook Messenger.
     */
    private final int facebookMessagingResponseTimeoutMillis;

    /**
     * The session timeout in milliseconds of the Facebook-Teneo bridge.
     */
    private final int bridgeSessionTimeoutMillis;

    /**
     * The maximal number of parallel session in the the Facebook-Teneo bridge.
     */
    private final Integer maxParallelSessions;

    /**
     * The Facebook Messenger webhook verification token.
     */
    private final String facebookWebhookVerificationToken;

    /**
     * The Facebook Messenger app secret.
     */
    private final String facebookAppSecret;

    /**
     * The mapping from Facebook page IDs to the access tokens of those pages.
     */
    private final Map<String, String> pageIdToAccessToken;

    /**
     * A flag indicating if some error and debug information should be added to requests sent both to
     * Teneo engine and to Facebook Messenger.
     */
    private final boolean explicitData;

    /**
     * Reads a properties file and creates a Properties object from its content.
     * 
     * @param file the properties file to read.
     * 
     * @return the Properties object.
     * 
     * @throws IOException if the file could not be read.
     */
    public static Properties getPropertiesFromFile(final File file) throws IOException {
        final Properties props = new Properties();
        try (final Reader reader = new FileReader(file, charset)) {
            props.load(reader);
        }
        return props;
    }

    /**
     * Creates an instance of this class from a properties file.
     * 
     * @param file the properties file to read.
     * 
     * @throws IOException if the file could not be read.
     * @throws IllegalArgumentException if the content of the file is not correct.
     */
    public Config(final File file) throws IOException, IllegalArgumentException {
        final Properties props = getPropertiesFromFile(file);
        final Logger logger = LoggerFactory.getLogger(Config.class);

        final Function<String, String> semiOblStr = p -> {
            String s = props.getProperty(p);
            if (s == null) throw new IllegalArgumentException("Missing obligatory non-empty property [" + p + "] in file [" + file + ']');
            if (logger.isDebugEnabled()) {
                if (s.isEmpty()) logger.debug("Reading config {} is empty", p);
                else logger.debug("Reading config {} = {}", p, s);
            }
            return s;
        };
        final Function<String, String> oblStr = p -> {
            String s = props.getProperty(p);
            if (s == null) throw new IllegalArgumentException("Missing obligatory non-empty property [" + p + "] in file [" + file + ']');
            if (s.isEmpty()) throw new IllegalArgumentException("Empty obligatory [" + p + "] in file [" + file + ']');
            if (logger.isDebugEnabled()) logger.debug("Reading config {} = {}", p, s);
            return s;
        };
        final ToIntFunction<String> oblNonnegativeIntProp = p -> {
            final String s = oblStr.apply(p);
            final int y;
            try {
                y = Integer.parseInt(s);
            } catch (final Exception ex) {
                throw new IllegalArgumentException("Bad integer [" + s + "] in property [" + p + "] in file [" + file + ']', ex);
            }
            if (y < 0) throw new IllegalArgumentException("Illegal negative integer [" + y + "] in property [" + p + "] in file [" + file + ']');
            return y;
        };
        final Function<String, URI> createUri = s -> {
            try {
                final URI uri = new URL(s).toURI();
                logger.debug("Creating URI [{}]", uri);
                return uri;
            } catch (final Exception ex) {
                throw new IllegalArgumentException("Bad URL/URI [" + s + "] in file [" + file + ']', ex);
            }
        };

        String s = oblStr.apply(PROP_TENEO_ENGINE_URL);
        teneoEngineEndpointUri = createUri.apply(s);
        if (teneoEngineEndpointUri.getPath() !=null && teneoEngineEndpointUri.getPath().indexOf(';') != -1) {
            throw new IllegalArgumentException("Bad URL/URI [" + teneoEngineEndpointUri + "] in file [" + file + "]: its path contains a semicolon");
        }
        teneoEngineConnectTimeoutMillis = oblNonnegativeIntProp.applyAsInt(PROP_TENEO_ENGINE_CONNECT_TIMEOUT);
        teneoEngineResponseTimeoutMillis = oblNonnegativeIntProp.applyAsInt(PROP_TENEO_ENGINE_RESPONSE_TIMEOUT);
        facebookMessagingConnectTimeoutMillis = oblNonnegativeIntProp.applyAsInt(PROP_FACEBOOK_MESSAGING_CONNECT_TIMEOUT);
        facebookMessagingResponseTimeoutMillis = oblNonnegativeIntProp.applyAsInt(PROP_FACEBOOK_MESSAGING_RESPONSE_TIMEOUT);
        bridgeSessionTimeoutMillis = oblNonnegativeIntProp.applyAsInt(PROP_BRIDGE_SESSION_TIMEOUT);
        maxParallelSessions = oblNonnegativeIntProp.applyAsInt(PROP_BRIDGE_MAX_PARALLEL_SESSIONS);
        facebookWebhookVerificationToken = oblStr.apply(PROP_FACEBOOK_WEBHOOK_VERIFICATION_TOKERN);
        facebookAppSecret = semiOblStr.apply(PROP_FACEBOOK_APPLICATION_SECRET);

        try (final JsonReader jsonReader = Json.createReader(new StringReader(oblStr.apply(PROP_FACEBOOK_PAGE_ID_TO_ACCESS_TOKEN)))) {
            final Map<String, JsonValue> m = jsonReader.readObject();
            pageIdToAccessToken = new HashMap<>(m.size());
            for (final Map.Entry<String, JsonValue> e : m.entrySet()) {
                final JsonValue v = e.getValue();
                if (v.getValueType() != JsonValue.ValueType.STRING) {
                    throw new IllegalArgumentException("Value type [" + v.getValueType() + "] in property [" + PROP_FACEBOOK_PAGE_ID_TO_ACCESS_TOKEN + "], only strings are allowed");
                }
                pageIdToAccessToken.put(e.getKey(), ((JsonString)v).getString());
            }
        }
        explicitData = Boolean.parseBoolean(props.getProperty(PROP_APPLICATION_EXPLICIT_DATA));
    }

    /**
     * Gets the URL (URI) of the Teneo engine.
     * 
     * @return the URL (URI) of the Teneo engine.
     * 
     * @see #PROP_TENEO_ENGINE_URL
     */
    public URI getTeneoEngineEndpointUri() {
        return teneoEngineEndpointUri;
    }

    /**
     * Gets the connection timeout of the requests sent to the Teneo engine.
     * 
     * @return the connection timeout of the requests sent to the Teneo engine.
     * 
     * @see #PROP_TENEO_ENGINE_CONNECT_TIMEOUT
     */
    public int getTeneoEngineConnectTimeoutMillis() {
        return teneoEngineConnectTimeoutMillis;
    }

    /**
     * Gets the response timeout of the requests sent to the Teneo engine.
     * 
     * @return the response timeout of the requests sent to the Teneo engine.
     * 
     * @see #PROP_TENEO_ENGINE_RESPONSE_TIMEOUT
     */
    public int getTeneoEngineResponseTimeoutMillis() {
        return teneoEngineResponseTimeoutMillis;
    }

    /**
     * Gets the connection timeout of the requests sent to Facebook Messenger.
     * 
     * @return the connection timeout of the requests sent to Facebook Messenger.
     * 
     * @see #PROP_FACEBOOK_MESSAGING_CONNECT_TIMEOUT
     */
    public int getFacebookMessagingConnectTimeoutMillis() {
        return facebookMessagingConnectTimeoutMillis;
    }

    /**
     * Gets the response timeout of the requests sent to Facebook Messenger.
     * 
     * @return the response timeout of the requests sent to Facebook Messenger.
     * 
     * @see #PROP_FACEBOOK_MESSAGING_RESPONSE_TIMEOUT
     */
    public int getFacebookMessagingResponseTimeoutMillis() {
        return facebookMessagingResponseTimeoutMillis;
    }

    /**
     * Gets the Facebook Messenger webhook verification token.
     * 
     * @return the Facebook Messenger webhook verification token.
     * 
     * @see #PROP_FACEBOOK_WEBHOOK_VERIFICATION_TOKERN
     */
    public String getFacebookWebhookVericationToken() {
        return facebookWebhookVerificationToken;
    }

    /**
     * Gets the Facebook Messenger app secret.
     * 
     * @return the Facebook Messenger app secret.
     * 
     * @see #PROP_FACEBOOK_APPLICATION_SECRET
     */
    public String getFacebookAppSecret() {
        return facebookAppSecret;
    }

    /**
     * Gets the session timeout in milliseconds of the Facebook-Teneo bridge.
     * 
     * @return the session timeout in milliseconds of the Facebook-Teneo bridge.
     * 
     * @see #PROP_BRIDGE_SESSION_TIMEOUT
     */
    public int getBridgeSessionTimeoutMillis() {
        return bridgeSessionTimeoutMillis;
    }

    /**
     * Gets the maximal number of parallel session in the the Facebook-Teneo bridge.
     * 
     * @return the maximal number of parallel session in the the Facebook-Teneo bridge.
     * 
     * @see #PROP_BRIDGE_MAX_PARALLEL_SESSIONS
     */
    public Integer getMaxParallelSessions() {
        return maxParallelSessions;
    }

    /**
     * Gets the page access token for the given page ID.
     * 
     * @param pageId the page ID
     * 
     * @return the page access token if defined, {@code null} otherwise.
     * 
     * @see #PROP_FACEBOOK_PAGE_ID_TO_ACCESS_TOKEN
     */
    public String getPageAccessToken(final String pageId) {
        return pageIdToAccessToken.get(pageId);
    }

    /**
     * Checks if some error and debug information should be added to requests sent both to Teneo engine
     * and to Facebook Messenger.
     * 
     * @return {@code true} if this information should be added, {@code false} otherwise.
     * 
     * @see #PROP_APPLICATION_EXPLICIT_DATA
     */
    public boolean isExplicitData() {
        return explicitData;
    }
}
