package com.artificialsolutions.asolfbconnector.facebook;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.net.URI;
import java.net.URL;
import java.net.URLEncoder;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.nio.charset.Charset;
import java.time.Duration;
import java.util.Map;
import java.util.concurrent.CompletableFuture;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.artificialsolutions.asolfbconnector.common.Config;
import com.artificialsolutions.common.JsonUtils;

/**
 * A class to send messages to Facebook Messenger in accordance with its
 * <a href="https://developers.facebook.com/docs/messenger-platform/reference/send-api/">Send
 * API</a> version 12.0.
 */
public class FacebookMessagingClient implements Serializable {

    /**
     * Serial version ID for object (de)serialization.
     */
    private static final long serialVersionUID = -7933456414549851074L;

    /**
     * The charset used for encoding.
     */
    private static final Charset charset = Charset.forName("UTF-8");

    /**
     * The content type header to be used in requests to Facebook Messaging.
     */
    private static final String requestContentTypeHeader = "application/json;charset=" + charset;

    /**
     * The configuration object.
     */
    private final Config config;

    /**
     * The URI (URL) of the Facebook Messaging endpoint.
     */
    private final URI endpointUri;

    /**
     * The beginning part of the message payload. It looks like this:
     * 
     * <pre>
     * {
     *   "recipient": {
     *     "id": <i>userPsid</i>
     *   },
     *   "message":
     * </pre>
     * 
     * Here, <code><i>userPsid</i></code> is the page-scoped user ID of the user who is the recipient of
     * the messages. In the case of text messages, the full payload looks as follows:
     * 
     * <pre>
     * {
     *   "recipient": {
     *     "id": "<i>userPsid</i>"
     *   },
     *   "message": {
     *     "text": "<i>The text submitted to the user</i>"
     *   }
     * }
     * </pre>
     */
    private final String payloadBegin;

    /**
     * The end part of the message payload. It looks like this:
     * 
     * <pre>
     * }
     * </pre>
     */
    private final static String payloadEnd = "}";

    /**
     * Response timeout in milliseconds. It is defined as transient since its serialization is forbidden
     * by its implementation despite implementing Serializable.
     */
    private transient Duration responseTimeoutMillis;

    /**
     * The HTTP client to send requests to Facebook Messenger.
     */
    private transient HttpClient httpClient;

    /**
     * The logger.
     */
    private transient Logger logger;

    /**
     * Indicates if potentially GDPR-sensitive data may ({@code true}) or may not ({@code false}) be
     * logged. The value of this property should be {@code false} in production.
     */
    private transient boolean logSensitive;

    /**
     * The Facebook endpoint URL to send the messages to.
     */
    private static final String messageEndpoint = "https://graph.facebook.com/v13.0/me/messages?access_token=";

    /**
     * Constructs an instance of the client to send messages to Facebook Messenger.
     * 
     * @param config The configuration object.
     * @param pageAccessToken The page access token of the Facebook business page whose messenger is
     * connected to Teneo engine.
     * @param userPsid The page scoped user ID (PSID) of the message recipient, i.e. of the Facebook
     * user who is talking to the Teneo engine via Facebook Messenger.
     * 
     * @throws NullPointerException if any of the arguments is {@code null}.
     * @throws IllegalArgumentException if page access token or user PSID are empty.
     */
    public FacebookMessagingClient(final Config config, final String userPsid, final String pageAccessToken) throws NullPointerException, IllegalArgumentException {
        if (config == null) throw new NullPointerException("config==null");
        this.config = config;
        if (pageAccessToken.isEmpty()) throw new IllegalArgumentException("Empty pageAccessToken");
        if (userPsid.isEmpty()) throw new IllegalArgumentException("Empty userPsid");
        try {
            this.endpointUri = new URL(messageEndpoint + URLEncoder.encode(pageAccessToken, charset)).toURI();

            final String s1 = "{\"recipient\":{\"id\":\"", s2 = "\"},\"message\":";
            final StringBuilder sb = new StringBuilder(s1.length() + s2.length() + (int)(userPsid.length() * 1.5));
            this.payloadBegin = JsonUtils.append(sb.append(s1), userPsid).append(s2).toString();
        } catch (final Exception ex) {
            throw new RuntimeException("Internal error creating Facebook messaging URL/URI from [" + messageEndpoint + "] and pageAccessToken [" + pageAccessToken + ']', ex);
        }
        initTransient();
        logger.trace("An instance of {} is constructed with the endpoint {}", FacebookMessagingClient.class, this.endpointUri);
    }

    /**
     * Initializes the transient class members.
     */
    private void initTransient() {
        logger = LoggerFactory.getLogger(FacebookMessagingClient.class);
        logSensitive = logger.isDebugEnabled() || logger.isTraceEnabled();
        responseTimeoutMillis = Duration.ofMillis(config.getFacebookMessagingResponseTimeoutMillis());
        httpClient = HttpClient.newBuilder()
                .version(HttpClient.Version.HTTP_1_1)
                .followRedirects(HttpClient.Redirect.NORMAL)
                .connectTimeout(Duration.ofMillis(config.getFacebookMessagingConnectTimeoutMillis()))
                .build();
    }

    private CompletableFuture<String> sendPayloadAsync(final String payload) {
        final long millisStart = System.currentTimeMillis();
        if (logger.isDebugEnabled()) {
            String t = "Async request [" + millisStart + "] start, URI [" + endpointUri + ']';
            if (logSensitive) t += ", payload [" + getAbbreviated(payload) + ']';
            logger.debug(t);
        }
        final CompletableFuture<String> r = httpClient.sendAsync(createRequest(payload), HttpResponse.BodyHandlers.ofString()).thenApply((final HttpResponse<String> response) -> {
            final String s = response.body();
            if (logger.isDebugEnabled()) {
                String t = "Async request [" + millisStart + "] end, URI [" + endpointUri + "], duration " + (System.currentTimeMillis() - millisStart) + " millisecs, status code " + response.statusCode() + ", headers [" + response.headers() + ']';
                if (logSensitive) t += ", payload [" + getAbbreviated(payload) + "], response body [ " + getAbbreviated(s) + " ]";
                logger.debug(t);
            }
            if (response.statusCode() != 200) {
                String t = "Async request [" + millisStart + "] error, URI [" + endpointUri + "], status code " + response.statusCode() + ", headers [" + response.headers() + ']';
                if (logSensitive) t += ", payload [" + getAbbreviated(payload) + "], response body [ " + getAbbreviated(s) + " ]";
                logger.error(t);
                throw new RuntimeException("Async request error, status code " + response.statusCode());
            }
            return s;
        });
        if (logger.isDebugEnabled()) {
            return r.whenComplete((final String x, final Throwable thr) -> {
                String t;
                if (thr != null) {
                    t = "Async request [" + millisStart + "] error, URI [" + endpointUri + "], duration " + (System.currentTimeMillis() - millisStart) + " millisecs";
                    if (logSensitive) t += ", payload [" + getAbbreviated(payload) + ']';
                    logger.warn(t, thr);
                } else {
                    t = "Async request [" + millisStart + "] success, URI [" + endpointUri + "], duration " + (System.currentTimeMillis() - millisStart) + " millisecs";
                    if (logSensitive) t += ", payload [" + getAbbreviated(payload) + "], result [ " + x + " ]";
                    logger.debug(t);
                }
            });
        }
        return r;
    }

    /**
     * Asynchronously sends a text to Facebook Messenger so it is then displayed to the user.
     * 
     * @param text the text to send.
     * 
     * @return a {@code java.util.concurrent.CompletableFuture<String>} object containing the "raw"
     * unparsed response of the Facebook Messenger platform. This response is NOT the user's response,
     * but some technical data indicating how the request has been processed by Facebook.
     * 
     * @see java.util.concurrent.CompletableFuture
     */
    public CompletableFuture<String> sendTextAsync(final String text) {
        final String payload = createTextPayload(text);
        return sendPayloadAsync(payload);
    }

    /**
     * Asynchronously sends a stringified JSON message to Facebook Messenger so it is then displayed to
     * the user.
     * 
     * @param stringifiedJsonMessage the stringified JSON message to send.
     * 
     * @return a {@code java.util.concurrent.CompletableFuture<String>} object containing the "raw"
     * unparsed response of the Facebook Messenger platform. This response is NOT the user's response,
     * but some technical data indicating how the request has been processed by Facebook.
     * 
     * @see java.util.concurrent.CompletableFuture
     */
    public CompletableFuture<String> sendAsync(final String stringifiedJsonMessage) {
        final String payload = createMessagePayload(stringifiedJsonMessage);
        return sendPayloadAsync(payload);
    }

    /**
     * Asynchronously sends a JSON-formatted message to Facebook Messenger so it is then displayed to
     * the user.
     * 
     * @param jsonMessage the JSON-formatted message to send.
     * 
     * @return a {@code java.util.concurrent.CompletableFuture<String>} object containing the "raw"
     * unparsed response of the Facebook Messenger platform. This response is NOT the user's response,
     * but some technical data indicating how the request has been processed by Facebook.
     * 
     * @see java.util.concurrent.CompletableFuture
     */
    public CompletableFuture<String> sendAsync(final Map<?, ?> jsonMessage) {
        final String payload = createMessagePayload(jsonMessage);
        return sendPayloadAsync(payload);
    }

    /**
     * Creates a request object with the given payload.
     * 
     * @param payload the payload for the request.
     * 
     * @return the created request.
     */
    private HttpRequest createRequest(final String payload) {
        return HttpRequest.newBuilder()
                .uri(endpointUri)
                .timeout(responseTimeoutMillis)
                .header("Content-Type", requestContentTypeHeader)
                .POST(HttpRequest.BodyPublishers.ofString(payload))
                .build();
    }

    /**
     * Creates request payload to send a test message.
     * 
     * @param text the text to be used in the payload.
     * 
     * @return the resulting payload.
     */
    private String createTextPayload(final String text) {
        final StringBuilder z = new StringBuilder(text.length() * 2 + payloadBegin.length() + payloadEnd.length() + 16);
        z.append(payloadBegin).append("{\"text\":\"");
        JsonUtils.append(z, text);
        z.append("\"}").append(payloadEnd);
        return z.toString();
    }

    /**
     * Creates request payload to send any message which is expected to be a stringified JSON object.
     * 
     * @param stringifiedJsonMessage the tringified JSON object to be used in the payload.
     * 
     * @return the resulting payload.
     */
    private String createMessagePayload(final String stringifiedJsonMessage) {
        return new StringBuilder(stringifiedJsonMessage.length() + payloadBegin.length() + payloadEnd.length())
                .append(payloadBegin).append(stringifiedJsonMessage).append(payloadEnd).toString();
    }

    /**
     * Creates request payload to send any message t.
     * 
     * @param jsonMessage the message object to be used in the payload.
     * 
     * @return the resulting payload.
     */
    private String createMessagePayload(final Map<?, ?> jsonMessage) {
        final StringBuilder z = new StringBuilder();
        z.append(payloadBegin).append("{\"text\":\"");
        JsonUtils.appendAsJson(z, (Map<?, ?>)jsonMessage);
        z.append(payloadEnd);
        return z.toString();
    }

    /**
     * Abbreviates a long text, to be displayed in logs etc.
     * 
     * @param s the text to abbreviate
     * 
     * @return the abbreviated text.
     */
    private static String getAbbreviated(final String s) {
        return s.length() > 256 ? s.substring(0, 253) + "..." : s;
    }

    /**
     * Reconstitutes this object from a stream (i.e., deserializes it).
     * 
     * @param ois the stream.
     * 
     * @throws IOException if the class of a serialized object could not be found.
     * @throws ClassNotFoundException if an I/O error occurs.
     */
    private void readObject(final ObjectInputStream ois) throws IOException, ClassNotFoundException {
        ois.defaultReadObject();
        initTransient();
        logger.trace("An instance of {} is created from deserialization", FacebookMessagingClient.class);
    }
}
