package com.artificialsolutions.asolfbconnector.webapp.servlets;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.artificialsolutions.asolfbconnector.bridge.FacebookTeneoBridge;
import com.artificialsolutions.asolfbconnector.common.Config;

import jakarta.json.Json;
import jakarta.json.JsonNumber;
import jakarta.json.JsonObject;
import jakarta.json.JsonReader;
import jakarta.json.JsonString;
import jakarta.json.JsonStructure;
import jakarta.json.JsonValue;
import jakarta.json.stream.JsonParsingException;
import jakarta.servlet.ServletException;
import jakarta.servlet.http.HttpServlet;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;

/**
 * Servlet implementing the
 * <a href="https://developers.facebook.com/docs/messenger-platform/webhook/">webhooks</a> logic of
 * <a href="https://developers.facebook.com/docs/messenger-platform">Facebook Messenger
 * Platform</a>. The current implementation has been developed and tested for the Facebook Graph API
 * version 12.0 and supports webhooks verification requests as well as <code>messages</code> and
 * <code>messaging_postbacks</code> webhooks. Requests coming from these webhooks (which are
 * triggered by users' interaction with the Facebook Messenger) are translated into corresponding
 * Teneo engine requests. Other webhooks don't trigger any Teneo engine request. And even though
 * they don't produce any error either, they are logged as warnings so it is possible to verify if
 * this application has accidentally been subscribed to some hooks it doesn't support. Teneo engine
 * responses are then submitted to the Facebook Messaging API so the Facebook Messenger displays
 * them to the user.
 */
public class Webhook extends HttpServlet {

    /**
     * Serial version ID for object (de)serialization.
     */
    private static final long serialVersionUID = 3341081028457194666L;

    /**
     * The configuration object.
     */
    private Config config;

    /**
     * The Facebook-Teneo bridge providing communication between both backends.
     */
    private FacebookTeneoBridge facebookTeneoBridge;

    /**
     * The logger.
     */
    private transient Logger logger;

    /**
     * Indicates if potentially GDPR-sensitive data may ({@code true}) or may not ({@code false}) be
     * logged. The value of this property should be {@code false} in production.
     */
    private transient boolean logSensitive;

    /*
     * The Facebook app secret in bytes.
     */
    private byte[] appSecretBytes;

    /**
     * Creates an instance of this class.
     */
    public Webhook() {
        initTransient();
    }

    /**
     * Initializes transient class members.
     */
    private void initTransient() {
        logger = LoggerFactory.getLogger(Webhook.class);
        logSensitive = logger.isDebugEnabled() || logger.isTraceEnabled();
    }

    /**
     * Checks if the security verification of the POST request originated by a Facebook webhook should
     * be done.
     * 
     * @return {@code true} if the security verification should be done, {@code false} otherwise.
     */
    private boolean isAppSecretSecurityVertificationActive() {
        return appSecretBytes != null;
    }

    /**
     * Encodes a String with HmacSHA1 using the app secret.
     * 
     * @param str the string to encode.
     * 
     * @return the input String encoded with HmacSHA1 using the app secret.
     * 
     * @throws Exception if the encoding failed.
     */
    private String hashHmac(final String str) throws Exception {
        Mac mac = Mac.getInstance("HmacSHA1");
        mac.init(new SecretKeySpec(appSecretBytes, "HmacSHA1"));
        byte[] hashBytes = mac.doFinal(str.getBytes());
        mac = null;
        int k = 0;
        final char[] res = new char[hashBytes.length * 2];
        for (final byte x : hashBytes) {
            res[k++] = Character.toLowerCase(Character.forDigit((x >> 4) & 0xF, 16));
            res[k++] = Character.toLowerCase(Character.forDigit((x & 0xF), 16));
        }
        hashBytes = null;
        return String.valueOf(res);
    }

    /**
     * Obtains a String payload from the request and closes the underlying reader.
     * 
     * @param request the request object.
     * 
     * @return the payload as a String.
     * 
     * @throws Exception if the underlying reader could not be accessed or reading the data failed.
     */
    private String stringPayloadFromRequestReader(final HttpServletRequest request) throws Exception {
        final Reader reader = request.getReader();
        final StringBuilder sb = new StringBuilder();
        try {
            int c;
            while ((c = reader.read()) != -1) {
                sb.append((char)c);
            }
        } finally {
            try {
                reader.close();
            } catch (final Exception ex) {
                logger.warn("Webhook POST request failure closing Reader", ex);
            }
        }
        return sb.toString();
    }

    /**
     * Returns some data of a potentially suspicious request for its logging.
     * 
     * @param request the request object.
     * 
     * @return the data as a String.
     */
    private String getBadRequestDataAsString(final HttpServletRequest request) {
        final StringBuilder sb = new StringBuilder();
        // request.getRemoteHost(), request.getRemoteAddr(), request.getLocalAddr(),
        // request.getServerName(), request.getLocalName()
        sb.append("RemoteHost: ").append(request.getRemoteHost());
        sb.append("; RemoteAddr: ").append(request.getRemoteAddr());
        sb.append("; LocalAddr: ").append(request.getLocalAddr());
        sb.append("; ServerName: ").append(request.getServerName());
        sb.append("; LocalName: ").append(request.getLocalName());
        sb.append("; headers:");
        for (final Enumeration<String> names = request.getHeaderNames(); names.hasMoreElements();) {
            final String name = names.nextElement();
            sb.append("\n[").append(name).append(": ");
            final Enumeration<String> values = request.getHeaders(name);
            if (values == null) sb.append("null");
            else {
                sb.append('[');
                boolean added = false;
                while (values.hasMoreElements()) {
                    if (added) sb.append(", ");
                    else added = true;
                    sb.append(values.nextElement());
                }
                sb.append(']');
            }
            sb.append(']');
        }
        return sb.toString();
    }

    /**
     * Processes regular Webhook POST requests.
     * 
     * @param request the request received from the Facebook messenger webhook API.
     * @param response the response returned to the Facebook messenger webhook API. Its MIME type is
     * <code>text/plain</code> and its encoding is UTF-8. If the request has been processed
     * successfully, the response will contain the response code {@code 200}, as per the Facebook
     * requirements. Otherwise the response code will reflect the problem encountered. The body of the
     * response will contain either the word OK or a short textual description of the problem.
     * 
     * @throws IOException {@inheritDoc}
     * @throws ServletException {@inheritDoc}
     */
    @Override
    protected void doPost(final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {
        response.setContentType("text/plain;charset=UTF-8");

        JsonReader jsonReader;

        if (isAppSecretSecurityVertificationActive()) {
            final String sigHeaderValue = request.getHeader("X-Hub-Signature");
            if (sigHeaderValue == null || sigHeaderValue.isEmpty()) {
                logger.error("Webhook POST payload has no X-Hub-Signature header value; " + getBadRequestDataAsString(request));
                response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Webhook POST payload has no X-Hub-Signature header value");
                return;
            }
            if (sigHeaderValue.startsWith("sha1=")) {
                try {
                    final String payload = stringPayloadFromRequestReader(request);
                    final String payloadSig = hashHmac(payload);
                    if (sigHeaderValue.regionMatches("sha1=".length(), payloadSig, 0, payloadSig.length())) {
                        logger.trace("Webhook POST payload has been successfully validated with signature [{}]", payloadSig);
                        jsonReader = Json.createReader(new StringReader(payload));
                    } else {
                        logger.error("Webhook POST payload has bad signature; payloadSig: [" + payloadSig + "], sigHeaderValue: [" + sigHeaderValue + "], payload: " + payload + ", " + getBadRequestDataAsString(request));
                        response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Webhook POST payload has bad signature");
                        return;
                    }
                } catch (final Exception ex) {
                    logger.error("Webhook POST failure verifying the payload signature; " + getBadRequestDataAsString(request));
                    response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Webhook POST failure verifying the payload signature");
                    return;
                }
            } else {
                logger.error("Webhook POST payload has bad X-Hub-Signature header value; " + getBadRequestDataAsString(request));
                response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Webhook POST payload has bad X-Hub-Signature header value");
                return;
            }
        } else {
            try {
                jsonReader = Json.createReader(request.getReader());
            } catch (final Exception ex) {
                logger.error("Webhook POST request failure obtaining reader", ex);
                response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Webhook POST request failure obtaining reader:\n" + ex);
                return;
            }
        }

        JsonStructure json;

        // https://developers.facebook.com/docs/messenger-platform/reference/webhook-events/messages
        // Receiving a JSON object from the webhook and parsing it:
        try {
            json = jsonReader.read();
        } catch (final JsonParsingException ex) {
            logger.error("Webhook POST request failure parsing JSON", ex);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, "Webhook POST request failure parsing JSON:\n" + ex);
            return;
        } catch (final Exception ex) {
            logger.error("Webhook POST request failure processing JSON", ex);
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Webhook POST request failure processing JSON:\n" + ex);
            return;
        } finally {
            try {
                jsonReader.close();
            } catch (final Exception ex) {
                logger.warn("Webhook POST request failure closing jsonReader", ex);
            }
            jsonReader = null;
        }
        final long reqTimePoint = System.currentTimeMillis();

        if (logSensitive && logger.isDebugEnabled()) logger.debug("Webhook POST request [{}]: {}", reqTimePoint, json);

        if (!(json instanceof JsonObject)) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, bad received JSON object type " + json.getClass().getName() + ": " + json : "Webhook POST request [" + reqTimePoint + "] failure, bad received JSON object type " + json.getClass().getName();
            logger.error(s);
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, s);
            return;
        }

        int totalEntries = 0, totalMessages = 0;

        // Obtain and verify the value of the "object" property, which should have the string value "page":
        JsonValue jv = ((JsonObject)json).get("object");
        if (jv == null) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, [object] property missing: " + json : "Webhook POST request [" + reqTimePoint + "] failure, [object] property missing";
            logger.error(s);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, s);
            return;
        }
        if (jv.getValueType() != JsonValue.ValueType.STRING) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, [object] property must have string value, has " + jv.getValueType() + ": " + json : "Webhook POST request [" + reqTimePoint + "] failure, [object] property must have string value, has " + jv.getValueType();
            logger.error(s);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, s);
            return;
        }
        if (!"page".equals(((JsonString)jv).getString())) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, [object] property must have string value [page], has [" + ((JsonString)jv).getString() + "]: " + json : "Webhook POST request [" + reqTimePoint + "] failure, [object] property must have string value [page], has [" + ((JsonString)jv).getString() + ']';
            logger.error(s);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, s);
            return;
        }

        // Obtain and verify the value of the "entry" property, which should be an array:
        if ((jv = ((JsonObject)json).get("entry")) == null) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, [entry] property missing: " + json : "Webhook POST request [" + reqTimePoint + "] failure, [entry] property missing";
            logger.error(s);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, s);
            return;
        }
        if (jv.getValueType() != JsonValue.ValueType.ARRAY) {
            final String s = logSensitive ? "Webhook POST request [" + reqTimePoint + "] failure, [entry] property must have array value, has " + jv.getValueType() + ": " + json : "Webhook POST request [" + reqTimePoint + "] failure, [entry] property must have array value, has " + jv.getValueType();
            logger.error(s);
            response.sendError(HttpServletResponse.SC_BAD_REQUEST, s);
            return;
        }

        final ArrayList<Map<String, String>> teneoInputs = new ArrayList<Map<String, String>>();
        final ArrayList<String> userPsids = new ArrayList<>(), pageAccessTokens = new ArrayList<>();

        // Iterating thru the values of the "entry" array:
        for (final JsonValue e : jv.asJsonArray()) {
            if (e == null) {
                logger.warn("Webhook POST request [{}], unexpected null element of [entry[]], {} was expected", reqTimePoint, JsonValue.ValueType.OBJECT);
                continue;
            }
            totalEntries++;

            if (e.getValueType() != JsonValue.ValueType.OBJECT) {
                logger.warn("Webhook POST request [{}], unexpected value type {} of an [entry[]] element, {} was expected", reqTimePoint, e.getValueType(), JsonValue.ValueType.OBJECT);
                continue;
            }
            final JsonObject individualEntry = e.asJsonObject();

            final String entryId;
            if ((jv = individualEntry.get("id")) == null || jv.getValueType() == JsonValue.ValueType.NULL) entryId = null;
            else if (jv.getValueType() == JsonValue.ValueType.STRING) entryId = ((JsonString)jv).getString();
            else {
                if (logSensitive) {
                    logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [id] property is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING, individualEntry);
                } else {
                    logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [id] property is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING);
                }
                continue;
            }

            final String asStandby;
            if ((jv = individualEntry.get("messaging")) == null || jv.getValueType() == JsonValue.ValueType.NULL) {
                if ((jv = individualEntry.get("standby")) == null || jv.getValueType() == JsonValue.ValueType.NULL) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [(messaging|standby)] property is missing: {}", reqTimePoint, individualEntry);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [(messaging|standby)] property is missing", reqTimePoint);
                    }
                    continue;
                }
                asStandby = "true";
            } else {
                asStandby = null;
            }

            if (jv.getValueType() != JsonValue.ValueType.ARRAY) {
                if (logSensitive) {
                    logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [(messaging|standby)] property is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.ARRAY, individualEntry);
                } else {
                    logger.warn("Webhook POST request [{}], skipping an [entry[]] element because its [(messaging|standby)] property is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.ARRAY);
                }
                continue;
            }

            // Iterating thru messaging / standby:
            for (final JsonValue m : jv.asJsonArray()) {
                if (m == null) {
                    logger.warn("Webhook POST request [{}], unexpected null element of [entry[].(messaging|standby)[]], {} was expected", reqTimePoint, JsonValue.ValueType.OBJECT);
                    continue;
                }
                totalMessages++;

                if (m.getValueType() != JsonValue.ValueType.OBJECT) {
                    logger.warn("Webhook POST request [{}], unexpected value type {} of an [entry[].(messaging|standby)[]] element, {} was expected", reqTimePoint, m.getValueType(), JsonValue.ValueType.OBJECT);
                    continue;
                }

                final JsonObject messagingElement = m.asJsonObject();

                if ((jv = messagingElement.get("account_linking")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_account_linking] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("delivery")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [message_deliveries] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("game_play")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_game_plays] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("pass_thread_control")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_handovers] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("optin")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_optins] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("policy_enforcement")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_policy_enforcement] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("reaction")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [message_reactions] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("read")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [message_reads] webhook is skipped", reqTimePoint);
                    continue;
                }
                if ((jv = messagingElement.get("referral")) != null && jv.getValueType() != JsonValue.ValueType.NULL) {
                    logger.warn("Webhook POST request [{}], a non implemented [messaging_referrals] webhook is skipped", reqTimePoint);
                    continue;
                }

                final JsonObject message;
                if ((jv = messagingElement.get("message")) == null || jv.getValueType() == JsonValue.ValueType.NULL) message = null;
                else if (jv.getValueType() == JsonValue.ValueType.OBJECT) message = jv.asJsonObject();
                else {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [message] property is of type {}, should be {}, null or absent: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [message] property is of type {}, should be {}, null or absent", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT);
                    }
                    continue;
                }
                final JsonObject postback;
                if ((jv = messagingElement.get("postback")) == null || jv.getValueType() == JsonValue.ValueType.NULL) postback = null;
                else if (jv.getValueType() == JsonValue.ValueType.OBJECT) postback = jv.asJsonObject();
                else {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [postback] property is of type {}, should be {}, null or absent: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [postback] property is of type {}, should be {}, null or absent", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT);
                    }
                    continue;
                }

                if (message == null && postback == null) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because both its [message] and [postback] are missing possibly because of a new, unimplemented webhook: {}", reqTimePoint, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because both its [message] and [postback] are missing possibly because of a new, unimplemented webhook", reqTimePoint);
                    }
                    continue;
                }

                if (message != null && postback != null) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because both its [message] and [postback] are present: {}", reqTimePoint, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because both its [message] and [postback] are present", reqTimePoint);
                    }
                    continue;
                }

                final String pageId;
                // Here, pageId is the ID of the customer's FB business page whose FB messenger is connected to
                // Teneo engine.
                if ((jv = messagingElement.get("recipient")) != null) {
                    if (jv.getValueType() == JsonValue.ValueType.OBJECT) {
                        if ((jv = jv.asJsonObject().get("id")) != null) {
                            if (jv.getValueType() != JsonValue.ValueType.STRING) {
                                if (logSensitive) {
                                    logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [recipient.id] is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING, messagingElement);
                                } else {
                                    logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [recipient.id] is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING);
                                }
                                continue;
                            }
                            pageId = ((JsonString)jv).getString();
                            if (entryId != null && !pageId.equals(entryId)) {
                                if (logSensitive) {
                                    logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its recipientId [{}] and entryId [{}] are different: {}", reqTimePoint, pageId, entryId, messagingElement);
                                } else {
                                    logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its recipientId [{}] and entryId [{}] are different", reqTimePoint, pageId, entryId);
                                }
                                continue;
                            }
                        } else {
                            pageId = entryId;
                        }
                    } else {
                        if (logSensitive) {
                            logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [recipient] is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT, messagingElement);
                        } else {
                            logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element becauseñ its [recipient] is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT);
                        }
                        continue;
                    }
                } else {
                    pageId = entryId;
                }

                if (pageId == null || pageId.isEmpty()) {
                    logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its both recipientId and entryId are missing: {}", reqTimePoint, messagingElement);
                    continue;
                }

                final String pageAccessToken = config.getPageAccessToken(pageId);
                if (pageAccessToken == null) {
                    // It means this connector is being access from an unauthorized FB page or the page ID wasn't
                    // configured correctly
                    if (logSensitive) {
                        logger.error("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its pageId (recipientId, entryId) [{}] has no page access token associated to it in the application configuration: {}", reqTimePoint, pageId, messagingElement);
                    } else {
                        logger.error("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its pageId (recipientId, entryId) [{}] has no page access token associated to it in the application configuration", reqTimePoint, pageId);
                    }
                    continue;
                }

                if ((jv = messagingElement.get("sender")) == null) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender] property is missing: {}", reqTimePoint, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender] property is missing", reqTimePoint);
                    }
                    continue;
                }
                if (jv.getValueType() != JsonValue.ValueType.OBJECT) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender] property is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender] property is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.OBJECT);
                    }
                    continue;
                }
                if ((jv = jv.asJsonObject().get("id")) == null) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender.id] is missing: {}", reqTimePoint, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender.id] is missing", reqTimePoint);
                    }
                    continue;
                }
                if (jv.getValueType() != JsonValue.ValueType.STRING) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender.id] is of type {}, should be {}: {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING, messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an [entry[].(messaging|standby)[]] element because its [sender.id] is of type {}, should be {}", reqTimePoint, jv.getValueType(), JsonValue.ValueType.STRING);
                    }
                    continue;
                }
                final String userPsid = ((JsonString)jv).getString();
                // userPsid (User's page-scoped ID) is used as the identifier of the session,
                // which is not HttpSession, but an custom concept used in this application.

                if (message != null && (jv = message.get("is_echo")) != null) {
                    switch (jv.getValueType()) {
                        case TRUE:
                            logger.warn("Webhook POST request [{}], a non implemented [message_echoes] webhook is skipped", reqTimePoint);
                            continue;
                        case FALSE:
                        case NULL:
                            break;
                        case NUMBER:
                            if (((JsonNumber)jv).intValue() == 0) break;
                            logger.warn("Webhook POST request [{}], a non implemented [message_echoes] webhook is skipped", reqTimePoint);
                            continue;
                        case STRING:
                            if (((JsonString)jv).getString().isEmpty()) break;
                        default:
                            logger.warn("Webhook POST request [{}], a non implemented [message_echoes] webhook is skipped", reqTimePoint);
                            continue;
                    }
                }

                final JsonObject holder = (message != null ? message : postback);

                final Map<String, String> teneoParamNameToValue = new HashMap<>();

                addPropertyStringValue(teneoParamNameToValue, "userinput", holder, "text");
                addPropertyStringValue(teneoParamNameToValue, "payload", holder, "payload");
                addPropertyStringValue(teneoParamNameToValue, "title", holder, "title");
                addPropertyStringValue(teneoParamNameToValue, "attachments", holder, "attachments");
                addPropertyStringValue(teneoParamNameToValue, "referral", holder, "referral");
                addPropertyStringValue(teneoParamNameToValue, "quick_reply.payload", holder.get("quick_reply"), "quick_reply");
                if (asStandby != null) teneoParamNameToValue.put("asStandby", asStandby);

                if (teneoParamNameToValue.isEmpty()) {
                    if (logSensitive) {
                        logger.warn("Webhook POST request [{}], skipping an empty [entry[].(messaging|standby)[]] input element: {}", messagingElement);
                    } else {
                        logger.warn("Webhook POST request [{}], skipping an empty [entry[].(messaging|standby)[]] input element");
                    }
                } else {
                    teneoParamNameToValue.put("viewtype", "tieapi");
                    teneoParamNameToValue.put("channel", "Facebook");
                    teneoParamNameToValue.put("businessPageId", pageId);
                    if (postback != null) teneoParamNameToValue.put("command", "messagingPostback");
                    if (config.isExplicitData()) {
                        // If explicitData has been activated in the application config,
                        // Teneo engine will additionally receive userPsid.
                        // Following the principle "the less potentially PII-related and
                        // not strictly necessary info, the better", it is not advisable
                        // to include this data other then for testing purposes.
                        teneoParamNameToValue.put("userPsid", userPsid);
                    }
                    teneoInputs.add(teneoParamNameToValue);
                    userPsids.add(userPsid);
                    pageAccessTokens.add(pageAccessToken);
                }
            }
        }

        switch (teneoInputs.size()) {
            case 1:
                // One supported message has been received. Submit it asynchronously to the Teneo/FB bridge:
                if (logger.isDebugEnabled()) {
                    logger.debug("Webhook POST request [{}], totalEntries: {}, totalMessages: {}, submittableMessages: {}", reqTimePoint, totalEntries, totalMessages, teneoInputs.size());
                    if (logSensitive) logger.debug("Webhook POST request [{}], asynchronously submitting for userPsid [{}] Teneo input {}", reqTimePoint, userPsids.get(0), teneoInputs.get(0));
                    else logger.debug("Webhook POST request [{}], asynchronously submitting Teneo engine input", reqTimePoint);
                }
                facebookTeneoBridge.submitTeneoFacebookAsync(userPsids.get(0), pageAccessTokens.get(0), teneoInputs.get(0));
                break;
            case 0:
                // No supported messages have been received. This is possible if this webhook has been
                // triggered as a result of the user clicking likes or similar actiaons not implemented here.
                if (logger.isInfoEnabled()) {
                    if (logSensitive) {
                        logger.info("Webhook POST request [{}], no messages have been processed, this can be due to errors or nonimplemented webhooks; totalEntries: {}, totalMessages: {}, original JSON: {}", reqTimePoint, totalEntries, totalMessages, json);
                    } else {
                        logger.info("Webhook POST request [{}], no messages have been processed, this can be due to errors or nonimplemented webhooks; totalEntries: {}, totalMessages: {}", reqTimePoint, totalEntries, totalMessages);
                    }
                }
                break;
            default:
                // Multiple supported messages have been received. Submit them
                // sequentially in a separate thread to the Teneo/FB bridge:
                if (logger.isDebugEnabled()) {
                    logger.debug("Webhook POST request [{}], totalEntries: {}, totalMessages: {}, submittableMessages: {}", reqTimePoint, totalEntries, totalMessages, teneoInputs.size());
                }
                final Thread thread = new Thread() {
                    @Override
                    public void run() {
                        int i = 0;
                        do {
                            if (logger.isDebugEnabled()) {
                                if (logSensitive) logger.debug("Webhook POST request [{}], sequentially submitting for userPsid [{}] Teneo input {}", reqTimePoint, userPsids.get(i), teneoInputs.get(i));
                                else logger.debug("Webhook POST request [{}], sequentially submitting Teneo input", reqTimePoint);
                            }
                            if (!facebookTeneoBridge.submitTeneoFacebook(userPsids.get(i), pageAccessTokens.get(i), teneoInputs.get(i))) {
                                if (logger.isWarnEnabled()) {
                                    if (logSensitive) logger.warn("Webhook POST request [{}], an error occurred or the session limit has been reached in the middle of sending multiple Teneo inputs (index {} of {} items) for userPsid [{}] and teneoInput {}", reqTimePoint, i, teneoInputs.size(), userPsids.get(i), teneoInputs.get(i));
                                    else logger.warn("Webhook POST request [{}], an error occurred or the session limit has been reached in the middle of sending multiple Teneo inputs (index {} of {} items) ", reqTimePoint, i, teneoInputs.size());
                                }
                            }
                        } while (++i != teneoInputs.size());
                    }
                };
                try {
                    thread.setDaemon(true);
                } catch (final Exception ex) {
                    if (logger.isInfoEnabled()) logger.info("Webhook POST request [{}], failure setting the auxiliary request thread as daemon", reqTimePoint, ex);
                }
                thread.start();
        }

        try {
            response.setStatus(HttpServletResponse.SC_OK);
            final PrintWriter pw = response.getWriter();
            pw.write("OK");
            pw.flush();
            pw.close();
        } catch (final Exception ex) {
            logger.error("Webhook POST request [{}] failure writing output", reqTimePoint, ex);
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Webhook POST request failure writing output:\n" + ex);
        }
    }

    /**
     * Webhook verification request processor. See how how handle verification requests in the <a href=
     * "https://developers.facebook.com/docs/workplace/integrations/custom-integrations/webhooks/">webhooks
     * documentation</a>.
     * 
     * @param request the verification request received from the Facebook messenger webhook API.
     * @param response the response returned to the Facebook messenger webhook API. Its MIME type is
     * <code>text/plain</code> and its encoding is UTF-8. If the request has been processed
     * successfully, the response will contain the response code {@code 200} the challenge value in the
     * response body body. Otherwise the response code will reflect the problem encountered and the body
     * of the response will contain a short textual description of the problem.
     * 
     * @throws IOException {@inheritDoc}
     * @throws ServletException {@inheritDoc}
     */
    @Override
    protected void doGet(final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {
        response.setContentType("text/plain;charset=UTF-8");
        final String mode = request.getParameter("hub.mode");
        final String verifyToken = request.getParameter("hub.verify_token");
        final String challenge = request.getParameter("hub.challenge");
        if (logger.isTraceEnabled()) logger.trace("Webhook GET request: challenge [{}], mode [{}], verifyToken [{}]", challenge, mode, verifyToken);
        final String errMsg;
        if ("subscribe".equals(mode)) {
            if (config.getFacebookWebhookVericationToken().equals(verifyToken)) {
                errMsg = (challenge == null) ? "Webhook token verification failed: missing challenge" : null;
            } else if (verifyToken == null) {
                errMsg = "Webhook token verification failed: missing verifyToken";
            } else if (verifyToken.isEmpty()) {
                errMsg = "Webhook token verification failed: empty verifyToken";
            } else {
                errMsg = "Webhook token verification failed: wrong verifyToken";
            }
        } else {
            errMsg = "Webhook token verification failed: mode [" + mode + ']';
        }
        try {
            final PrintWriter pw = response.getWriter();
            if (errMsg == null) {
                logger.debug("Webhook token verification succeeded");
                response.setStatus(HttpServletResponse.SC_OK);
                pw.write(challenge);
            } else {
                logger.warn(errMsg);
                response.setStatus(HttpServletResponse.SC_FORBIDDEN);
                pw.write(errMsg);
            }
            pw.flush();
            pw.close();
        } catch (final Exception ex) {
            logger.error("Webhook GET request failure obtaining writer", ex);
            response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Webhook GET request failure obtaining writer:\n" + ex);
        }
    }

    private static String getPropertyStringValue(JsonValue x, final String propertyName) {
        if (x != null && x.getValueType() == JsonValue.ValueType.OBJECT && (x = ((JsonObject)x).get(propertyName)) != null) {
            switch (x.getValueType()) {
                case STRING:
                    final String s = ((JsonString)x).getString();
                    return s.isEmpty() ? null : s;
                case NULL:
                case FALSE:
                    return null;
                default:
                    return x.toString();
            }
        }
        return null;
    }

    private static void addPropertyStringValue(final Map<String, String> m, final String newPropertyName, final JsonValue holder, final String originalPropertyName) {
        final String s = getPropertyStringValue(holder, originalPropertyName);
        if (s != null) m.put(newPropertyName, s);
    }

    /**
     * Initializes this servlet.
     * 
     * @throws ServletException {@inheritDoc}
     */
    @Override
    public void init() throws ServletException {
        config = (Config)getServletContext().getAttribute("config");
        appSecretBytes = config.getFacebookAppSecret().isEmpty() ? null : config.getFacebookAppSecret().getBytes();
        facebookTeneoBridge = new FacebookTeneoBridge(config);
    }

    /**
     * Reconstitutes this servlet from a stream (i.e., deserializes it).
     * 
     * @param ois the stream.
     * 
     * @throws IOException if the class of a serialized object could not be found.
     * @throws ClassNotFoundException if an I/O error occurs.
     */
    private void readObject(final ObjectInputStream ois) throws IOException, ClassNotFoundException {
        ois.defaultReadObject();
        initTransient();
    }
}
